"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const debug_1 = require("../debug");
const debug = debug_1.default('pnpm:install');
const npa = require("npm-package-arg");
const fs = require("mz/fs");
const logger = require("@zkochan/logger");
const path = require("path");
const rimraf = require("rimraf-then");
const resolve_1 = require("../resolve");
const requireJson_1 = require("../fs/requireJson");
const exists = require("exists-file");
const isAvailable_1 = require("./isAvailable");
const memoize_1 = require("../memoize");
const fetchResolution_1 = require("./fetchResolution");
/**
 * Installs a package.
 *
 * What it does:
 *
 * - resolve() - resolve from registry.npmjs.org
 * - fetch() - download tarball into node_modules/.store/{name}@{version}
 * - recurse into its dependencies
 * - symlink node_modules/{name}
 *
 * @param {Object} ctx - the context.
 * @param {Object} pkgMeta - meta info about the package to install.
 *
 * @example
 *     install(ctx, 'rimraf@2', './node_modules')
 */
function fetch(ctx, pkgRawSpec, modules, options) {
    return __awaiter(this, void 0, void 0, function* () {
        debug('installing ' + pkgRawSpec);
        // Preliminary spec data
        // => { raw, name, scope, type, spec, rawSpec }
        const spec = npa(pkgRawSpec);
        // Dependency path to the current package. Not actually needed anmyore
        // outside getting its length
        // => ['babel-core@6.4.5', 'babylon@6.4.5', 'babel-runtime@5.8.35']
        const keypath = (options && options.keypath || []);
        const log = logger.fork(spec).log.bind(null, 'progress', pkgRawSpec);
        try {
            let resolution = ctx.shrinkwrap[pkgRawSpec];
            if (!resolution) {
                // it might be a bundleDependency, in which case, don't bother
                const available = !options.force && (yield isAvailable_1.default(spec, modules));
                if (available) {
                    const fetchedPkg = yield saveCachedResolution();
                    fetchedPkg.fetchingPkg.then(pkg => log('package.json', pkg));
                    fetchedPkg.fetchingFiles.then(() => log('done'));
                    return fetchedPkg;
                }
            }
            if (!resolution || options.update) {
                resolution = yield resolve_1.default(spec, {
                    log,
                    got: options.got,
                    root: options.root,
                    linkLocal: options.linkLocal,
                    tag: options.tag
                });
                if (resolution.tarball || resolution.repo) {
                    ctx.shrinkwrap[pkgRawSpec] = Object.assign({}, resolution);
                    delete ctx.shrinkwrap[pkgRawSpec].pkg;
                    delete ctx.shrinkwrap[pkgRawSpec].fetch;
                    delete ctx.shrinkwrap[pkgRawSpec].root;
                }
            }
            log('resolved', resolution);
            const target = path.join(options.storePath, resolution.id);
            const fetchingFiles = fetchToStoreCached({
                fetchLocks: ctx.fetchLocks,
                target,
                resolution,
                log,
                got: options.got,
                force: options.force,
            });
            const fetchingPkg = resolution.pkg
                ? Promise.resolve(resolution.pkg)
                : fetchingFiles.then(() => requireJson_1.default(path.join(target, 'package.json')));
            const fetchedPkg = {
                fetchingPkg,
                fetchingFiles,
                id: resolution.id,
                fromCache: false,
                path: target,
                srcPath: resolution.root,
                abort: function () {
                    return __awaiter(this, void 0, void 0, function* () {
                        try {
                            yield fetchingFiles;
                        }
                        finally {
                            return rimraf(target);
                        }
                    });
                },
            };
            fetchedPkg.fetchingPkg.then(pkg => log('package.json', pkg));
            fetchedPkg.fetchingFiles.then(() => log('done'));
            return fetchedPkg;
        }
        catch (err) {
            log('error', err);
            throw err;
        }
        function saveCachedResolution() {
            return __awaiter(this, void 0, void 0, function* () {
                const target = path.join(modules, spec.name);
                const stat = yield fs.lstat(target);
                if (stat.isSymbolicLink()) {
                    const linkPath = yield fs.readlink(target);
                    return save(path.resolve(linkPath, target));
                }
                return save(target);
                function save(fullpath) {
                    return __awaiter(this, void 0, void 0, function* () {
                        const data = yield requireJson_1.default(path.join(fullpath, 'package.json'));
                        return {
                            fetchingPkg: Promise.resolve(data),
                            fetchingFiles: Promise.resolve(),
                            id: path.basename(fullpath),
                            fromCache: true,
                            path: fullpath,
                            abort: () => Promise.resolve(),
                        };
                    });
                }
            });
        }
    });
}
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = fetch;
/**
 * Fetch to `.store/lodash@4.0.0`
 * If an ongoing build is already working, use it. Also, if that ongoing build
 * is part of the dependency chain (ie, it's a circular dependency), use its stub
 */
function fetchToStoreCached(opts) {
    return memoize_1.default(opts.fetchLocks, opts.resolution.id, function () {
        return __awaiter(this, void 0, void 0, function* () {
            if (!opts.force && (yield exists(opts.target)))
                return;
            yield rimraf(opts.target);
            opts.log('download-queued');
            yield fetchResolution_1.default(opts.resolution, opts.target, { got: opts.got, log: opts.log });
            const pkg = yield requireJson_1.default(path.join(opts.target, 'package.json'));
            opts.log('package.json', pkg);
        });
    });
}
//# sourceMappingURL=fetch.js.map